#ifndef __MPI_BINDINGS
#define __MPI_BINDINGS
#define _MPI_INCLUDE
//
// To do:  make non-static functions virtual.  Note that the standard
// requires this (see "Class Member Functions for MPI").

//

/* C++ information */
namespace MPI {

  // Basic types
  typedef MPI_Aint   Aint;
  typedef MPI_Offset Offset;
  typedef MPI_Fint   Fint;

  // Special implementation macros
#ifdef HAVE_CXX_EXCEPTION
  // This is not completely correct yet; we need to ensure that an error code 
  // is returned.  What do we do for other error handlers?
  // Should there be a special invocation switch?  Should we avoid inlining
  // to maintain separation from the user's code (we can then use the internal
  // structure of the objects within the implementation of the C++ wrappers,
  // much as we could within the Fortran wrappers.
  // An alternative is to have a function that gets the error behavior to
  // use before calling the C version of the routine; this routine can 
  // know details of the internal structures.
#define MPIX_CALL( fnc ) \
{int err; err = fnc; if (err) throw Exception(err);}
#else
#define MPIX_CALL( fnc ) fnc
#endif

  // MPI classes

  class Datatype {
    friend class Comm;
  protected:
    MPI_Datatype the_real_dtype;
  public:
    inline Datatype(MPI_Datatype dtype) { the_real_dtype = dtype; }
    inline Datatype(void) {the_real_dtype = MPI_DATATYPE_NULL;}
    inline operator MPI_Datatype*() { return &the_real_dtype; }
    inline operator MPI_Datatype() { return the_real_dtype; }
    Datatype& Datatype::operator=(const MPI_Datatype& the_real_dtype);
    Datatype :: Datatype(const Datatype &the_real_dtype);

    // MPI 1 functions
#include "mpi1cppdtype.h"
    // MPI 2 functions
#include "mpi2cppdtype.h"
  }; // Datatype

  class Info {
  protected:
    MPI_Info the_real_info;
  public:
    inline Info(MPI_Info info) { the_real_info = info; }
    inline Info(void) {the_real_info = MPI_INFO_NULL;}
    inline operator MPI_Info*() { return &the_real_info; }
    inline operator MPI_Info() { return the_real_info; }
    Info& Info::operator=(const MPI_Info& the_real_info);
    Info :: Info(const Info &the_real_info);
    // MPI 2 functions
#include "mpi2cppinfo.h"
  }; // Info

  class Status {
  protected: 
    MPI_Status the_real_status;
  public:
    inline Status(MPI_Status status) { the_real_status = status; }
    inline Status(void) {}
    inline operator MPI_Status() { return the_real_status; }
    Status& Status::operator=(const MPI_Status& the_real_status);
    inline operator MPI_Status*() { return &the_real_status; }
    Status :: Status(const Status &the_real_status);

    // MPI 1 functions
#include "mpi1cppst.h"
    // MPI 2 functions
#include "mpi2cppst.h"
  }; // Status

  class Group {
  protected:
    MPI_Group the_real_group;
  public:
    inline Group(MPI_Group group) { the_real_group = group; }
    inline Group(void) {the_real_group = MPI_GROUP_NULL;}
    inline operator MPI_Group*() { return &the_real_group; }
    inline operator MPI_Group() { return the_real_group; }
    Group& Group::operator=(const MPI_Group& the_real_group);
    Group :: Group(const Group &the_real_group);

    // MPI 1 functions
#include "mpi1cppgroup.h"
  }; // Group

  class Op {
  protected:
    MPI_Op the_real_op;
  public:
    inline Op(MPI_Op op) { the_real_op = op; }
    inline Op(void) {the_real_op = MPI_OP_NULL;}
    inline operator MPI_Op*() { return &the_real_op; }
    inline operator MPI_Op() { return the_real_op; }
    Op& Op::operator=(const MPI_Op& the_real_op);
    Op :: Op(const Op &the_real_op);

    // MPI 1 functions
    // A complication here is that the MPI implementation knows how to call
    // C routines, not C++ routines.
#include "mpi1cppop.h"
  }; // Op

  class Errhandler {
  protected:
    MPI_Errhandler the_real_errhandler;
  public:
    // MPI 1 functions
#include "mpi1cpperrh.h"
  }; // Errhandler

  class Exception { 
  protected:
    int the_real_code;
  public: 
    inline Exception(int error_code) { the_real_code = MPI_SUCCESS; }
    inline Exception(int err) { the_real_code = err; }
    inline ~Exception(){};

    int Get_error_code() const { return the_real_code; }; 
    int Get_error_class() const { 
      int err_class; MPI_Error_class( the_real_code, &err_class ); 
      return err_class; }; 
    const char *Get_error_string() const {
      char *str = new[MPI_MAX_ERROR_STRING];
      MPI_Error_string( the_real_code, str, &len ); return str;}; 
  }; 

  class Request {
  protected:
    MPI_Request the_real_req;
    // Need Comm for pack/unpack/packsize
    friend class Comm;
  public:
    inline Request(MPI_Request req) { the_real_req = req; }
    inline Request(void) {the_real_req = MPI_REQUEST_NULL;};
    inline operator MPI_Request*() { return &the_real_req; }
    inline operator MPI_Request() { return the_real_req; }
    Request& Request::operator=(const MPI_Request& the_real_req);
    Request :: Request( const Request &the_real_req );

    // MPI 1 functions
#include "mpi1cppreq.h"
    // MPI 2 functions
#include "mpi2cppreq.h"
  };  // Request

  class Prequest : public Request {
  protected:
    MPI_Request the_real_req;
  public:
    // MPI  1 functions
#include "mpi1cpppreq.h"
  }; // Prequest

  class Comm { 
  protected:
    MPI_Comm the_real_comm;

  public:
    inline Comm(void) { the_real_comm = MPI_COMM_NULL; }
    inline Comm(MPI_Comm comm) { the_real_comm = comm; }
    inline operator MPI_Comm*() { return &the_real_comm; }
    inline operator MPI_Comm() { return the_real_comm; }
    Comm& Comm::operator=(const MPI_Comm& the_real_comm);
    Comm :: Comm(const Comm &the_real_comm);

    // MPI 1 functions
#include "mpi1cppcomm.h"
    // MPI 2 functions
#include "mpi2cppcomm.h"
  };  // Comm

  class Intercomm: public Comm {
    // Need Comm for Merge
    friend class Intracomm;
  protected:
    MPI_Comm the_real_comm;
  public:
    // MPI 1 functions
#include "mpi1cppinter.h"
    // MPI 2 functions
#include "mpi2cppinter.h"
  }; // Intercomm

  class Intracomm : public Comm {
  public: 
    // For the create_cart/graph routines
    friend class Cartcomm;
    friend class Graphcomm;

    inline Intracomm(MPI_Comm comm) { the_real_comm = comm; }
    inline Intracomm(void) {the_real_comm = MPI_COMM_NULL;}
    //inline operator MPI_Comm*() { return &the_real_comm; }
    //inline operator MPI_Comm() { return the_real_comm; }
    // MPI 1 functions
#include "mpi1cppintra.h"
    // MPI 2 functions
#include "mpi2cppintra.h"
  }; // Intracomm


  class Grequest: public Request {
    // MPI 2 functions
#include "mpi2cppgreq.h"
  }; // Grequest;

#ifdef HAVE_MPI_WIN_CREATE
  class Win { 
  protected:
    MPI_Win the_real_win;
  public:
    inline Win(MPI_Win win) { the_real_win = win; }
    inline Win(void) {the_real_win = MPI_WIN_NULL;}
    inline operator MPI_Win*() { return &the_real_win; }
    inline operator MPI_Win() { return the_real_win; }
    Win& Win::operator=(const MPI_Win& the_real_win);
    Win :: Win(const Win &the_real_win);
    // MPI 2 functions
#include "mpi2cppwin.h"
  }; // Win;
#endif

  class File {
  protected:
    MPI_File the_real_file;
  public:
    inline File(MPI_File file) { the_real_file = file; }
    inline File(void) {the_real_file = MPI_FILE_NULL;}
    inline operator MPI_File*() { return &the_real_file; }
    inline operator MPI_File() { return the_real_file; }
    File& File::operator=(const MPI_File& the_real_file);
    File :: File(const File &the_real_file);
    // MPI 2 functions
#include "mpi2cppfile.h"
  }; // File;

  class Graphcomm : public Intracomm {
  protected:
    MPI_Comm the_real_comm;
  public:
    // MPI 1 functions
#include "mpi1cppgraph.h"
  }; // Graphcomm;

  class Cartcomm  : public Intracomm {
  protected:
    MPI_Comm the_real_comm;
  public:
    inline Cartcomm(void) {the_real_comm = MPI_COMM_NULL;}
    inline operator MPI_Comm*() { return &the_real_comm; }
    inline operator MPI_Comm() { return the_real_comm; }
    Cartcomm& Cartcomm::operator=(const MPI_Comm& the_real_comm);
    Cartcomm :: Cartcomm(const Comm &the_real_comm);
// MPI 1 functions
#include "mpi1cppcart.h"
  }; // Cartcomm;
  
  // General MPI namespace
  void MPI_CXX_Init( void );

  // Predefined communicators
  extern Intracomm COMM_WORLD;
  extern Intracomm COMM_SELF;

  // Predefined datatypes
  // Can these be const in MPI-2 (set_name?)
  extern const Datatype CHAR;
  extern const Datatype UNSIGNED_CHAR;
  extern const Datatype BYTE;
  extern const Datatype SHORT;
  extern const Datatype UNSIGNED_SHORT;
  extern const Datatype INT;
  extern const Datatype UNSIGNED;
  extern const Datatype LONG;
  extern const Datatype UNSIGNED_LONG;
  extern const Datatype FLOAT;
  extern const Datatype DOUBLE;
  extern const Datatype LONG_DOUBLE;
  extern const Datatype LONG_LONG_INT;
  extern const Datatype LONG_LONG;
  extern const Datatype UNSIGNED_LONG_LONG;
  extern const Datatype PACKED;
  extern const Datatype LB;
  extern const Datatype UB;
  extern const Datatype FLOAT_INT;
  extern const Datatype DOUBLE_INT;
  extern const Datatype LONG_INT;
  extern const Datatype SHORT_INT;
  extern const Datatype TWOINT;
  extern const Datatype LONG_DOUBLE_INT;

  // Functions part of MPI but no particular class

#include "mpi1cppgen.h"
#include "mpi2cppgen.h"

  // Static values and "execution-time" constants
  extern const Comm COMM_NULL;

  // Results of compare
  extern const int IDENT;
  extern const int CONGRUENT;
  extern const int SIMILAR;
  extern const int UNEQUAL;

  // Error Classes
  extern const int SUCCESS;
  extern const int ERR_BUFFER;
  //etc.

  // Point-to-point constants
  extern const int ANY_TAG;
  extern const int ANY_SOURCE;
  extern const int PROC_NULL;
  extern const int UNDEFINED;
  extern const int BSEND_OVERHEAD;

  // Misc constants
  extern const int KEYVAL_INVALID;

  // String sizes
  extern const int MAX_PROCESSOR_NAME;
  extern const int MAX_ERROR_STRING;
  //??  extern const int MAX_OBJECT_NAME;

};  // MPI namespace

#endif

